Attribute VB_Name = "Statistique"
Option Explicit
Option Base 0

Public Sub Regression(X() As Double)
'
'----------------------------------------------------------------------------------------
' Regression - Moindres carrs ordinaires avec constante
'----------------------------------------------------------------------------------------
'
' Ce programme de rgression est tir de :
'
'   Almon Clopper (1967), "Matrix in economics", Addison Wesley, Reading Massachusetts,
'   Chapiter 1-3, pp. 1-57. Le chapitre 3 contient une version explique du code Fotran.
'
'   Il s'agit ici d'une version crite en utilisant le langage de programmation
'   Visual Basic 6 de Microsoft.
'
'----------------------------------------------------------------------------------------
'
' NVI : Nombre de variables indpendantes (NIV number of independant variables in Almon)
' NC  : NVI + 1, colonne de X contenant la variable  explique.
' NO  : Nombre d'observations (idem dans Almon)
'
'----------------------------------------------------------------------------------------
'  2003 - Paul Salmon
'----------------------------------------------------------------------------------------
'
' Dclarations locales
'
Dim NVI As Long                ' Nombre de Variables Indpendantes
Dim NO As Long                 ' Nombre d'Observations
Dim NC As Long                 ' Nombre de Colonnes de la matrice X, (NVI + 1)
'
Dim NCR As Long                ' NVI rectifi
Dim NOR As Long                ' NO rectifi
Dim NVIR As Long               ' NVIR rectifi
'
' Scalaires et Matrice permettant d'effectuer l'inversion de la matrice X'X.
'
Dim A() As Double              ' Matrice contenant les rsultats des calculs
                               ' intermdiaires et terminal de la matrice inverse
                               ' de X'X augmente
Dim Pivot As Double            ' Valeur du pivot
Dim AJI As Double              '
'
' Vecteurs et Matrices temporaires pour la rcuprations de rsultats intermdiaires.
'
Dim Temp() As Double           '
Dim TempS() As Double          '
'
' Scalaires, Vecteurs et Matrices contenant les rsultats permettant d'analyser les
' rsultats obtenus par la mthode des moindres carrs ordinaires (MCO).
'
Dim Beta() As Double           ' Beta des coefficients
Dim Elasticity() As Double     ' Elasticit des coefficients
Dim DoFree As Long             ' Degr de libert
Dim DP() As Double             '
Dim DW As Double               ' Durbin-Watson
Dim errp1 As Double            '
Dim FTest                      ' Valeur du fisher pour la rgression
Dim lresid As Double           '
Dim MAPETEMP As Double         '
Dim MAPE As Double             ' Mean Absolute Percentage Error
Dim Mean() As Double           ' Moyenne de la variable  expliqur et des variables
                               ' explicatives.
Dim Mexval() As Double         ' MexVal
Dim nobs As Long               ' Nombre d'observations
Dim NorRes() As Double         '
Dim R2 As Double               ' R2
Dim RBSQR As Double            ' R2 corrig
Dim Resid() As Double          ' Rsidus
Dim RHO As Double              ' RHO
Dim s2 As Double               ' Ecart-type des rsidus
Dim SEE As Double              '
Dim SEEP1 As Double            '
Dim SigmaCoef() As Double      ' Ecart-Type des coefficients
Dim somme1 As Double           '
Dim SR2 As Double              ' Variance des rsidus
Dim SY2 As Double              ' Variance de Y
Dim tstudent() As Double       ' Tde student
Dim YHAT() As Double           ' Y calculs
'
' Indice de boucle
'
Dim i As Long                  ' Indice de boucle
Dim j As Long                  ' Indice de boucle
Dim k As Long                  ' Indice de boucle
Dim n As Long                  ' Indice de boucle
'
' Initialisation des variables
'   Cette initialisation doit tre faite avant le dimensionnement des
'   vecteurs et tableaux.
'
NC = UBound(X, 2)              '
NVI = NC - 1                   '
NO = UBound(X, 1)              '
'
NCR = NC + 1                   '
NVIR = NVI + 1                 '
NOR = NO + 1                   '
'
' Dimensionnement des Vecteurs et Tableaux
'
ReDim A(0 To NC, 0 To NC)                ' Matrice rsultant du produit de X'X "augmente"
ReDim Beta(0 To NVI)                     ' Vecteur des Beta calculs
ReDim Coef(0 To NVI, 0 To 8)             ' Vecteur des coeficients de rgression
ReDim DP(0 To NC, 0 To NC)               ' Matrice des drives partielles
ReDim Elasticity(0 To NVI)               ' Vecteur des lasticits clacules
ReDim MatCrossCorr(0 To NC, 0 To NC)     ' Matrice de corrlation croise
ReDim MatVarCovCoef(0 To NVI, 0 To NVI)  ' Matrice de Variance-covariance des coefficients
ReDim Mean(0 To NC)                      ' Vecteur des moyennes
ReDim Mexval(0 To NVI)                   ' Vecteur des Mexvals
ReDim NorRes(0 To NVI)                   ' Vecteurs des rsidus normaliss
ReDim PD(0 To NVI, 0 To NVI)             '
ReDim RegData(0 To UBound(X, 1), 0 To 3) ' Matrice contenant Y, Yhat et resid
ReDim Resid(0 To NO)                     ' Vecteur des rsidus
ReDim SigmaCoef(0 To NVI)                ' Vecteur des cart-type des coefficients
ReDim StatReg(0 To 10)                   ' Vecteur des rsultats statistiques
ReDim Temp(0 To NC, 0 To NC)             ' Matrice Temporaire
ReDim TempS(0 To NVI)                    ' Vecteur Temporaire
ReDim tstudent(0 To NVI)                 ' Vecteur des t-students calculs
ReDim YHAT(0 To NO)                      ' Vecteur des valeurs calcules de Y
'
'----------------------------------------------------------------------------------------
' Construction de la Matrice X'X augmente
'----------------------------------------------------------------------------------------
'
For i = 0 To NC
    For j = 0 To NC
        '
        ' Initialisation  0. de l'lment (i,j) de A
        '
        A(i, j) = 0#
        '
        ' Calcul de la matrice symtrique X'X
        '
        For n = 0 To NO
            A(i, j) = A(i, j) + X(n, i) * X(n, j)
        Next n
        '
        ' Remplissage de la partie infrieure de la matrice
        '
        A(j, i) = A(i, j)
        '
    Next j
Next i
'
'----------------------------------------------------------------------------------------
' Inversion de la matrice X'X "augmente"
'----------------------------------------------------------------------------------------
'
For i = 0 To NVI
    '
    ' Initialisation du pivot de la ligne
    '
    Pivot = A(i, i)
    '
    ' Division du pivot de la ligne par lui-mme
    '
    A(i, i) = 1
    '
    ' Division du pivot lment par le pivot ligne
    '
    For k = 0 To NC
        If Pivot <> 0 Then
            A(i, k) = A(i, k) / Pivot
        Else
            MsgBox "Le pivot est nul pour la variable " & i & ". Le programme est arrt."
            End
        End If
    Next k
    '
    ' Rduction des lignes "non pivot"
    '
    For j = 0 To NC
        '
        ' Sauter la ligne pivot
        '
        If i <> j Then
            '
            '
            '
            AJI = A(j, i)
            '
            '
            '
            A(j, i) = 0#
            '
            '
            '
            For k = 0 To NC
                A(j, k) = A(j, k) - A(i, k) * AJI
                '
                ' A revoir ....
                '
                ' Rcupration des valeurs des moyennes et des variances-covariances.
                '
                ' i = 0; premier pivot
                '
                If i = 0 Then
                    If k = 0 Then
                        Temp(0, k) = 1#
                        Temp(j, k) = -A(j, k)
                    Else
                        Temp(0, k) = 0#
                        Temp(j, k) = A(j, k)
                    End If
                End If
                '
                ' Rcupration de la Variance de Y
                '
                If (j = NC) And (k = NC) Then
                        TempS(i) = A(NC, NC)
                End If
                '
            Next k
        End If
    Next j
Next i
'
'----------------------------------------------------------------------------------------
' Calculs des diverses matrices et statistiques de diagnostic
'----------------------------------------------------------------------------------------
'
' Drives partielles de la rgression
'
For i = 0 To NVI
    For k = 0 To i
        For j = 0 To i
            PD(k, j) = A(j, k) / A(j, j)
        Next j
    Next k
Next i
'
' Matrice de corrlation
'
For i = 1 To NC
    For j = 1 To NC
        If i = j Then
            MatCrossCorr(i, j) = 1#
        Else
            MatCrossCorr(i, j) = Temp(i, j) / (Sqr(Temp(i, i) * Temp(j, j)))
        End If
    Next j
Next i
'
' Pour faciliter la lecture des formules.
'
SR2 = TempS(NVI)
SY2 = TempS(0)
'
' Standard Error of Estimate
'
' Cas Particulier suivant les directives de Clopper Almon.
'
SEE = Sqr(SR2 / NOR)
'
' Calcul du R2
'
R2 = 1 - (SR2 / SY2)
'
' Calcul RBSQ
'
RBSQR = 1 - ((SR2 / (NOR - NVIR)) / (SY2 / (NOR - 1)))
'
' Ecart-tye des rsidus
'
s2 = SR2 / (NOR - NVIR)
'
' Calcul de la matrice de Variance-covariance des coefficients
'
For i = 0 To NVI
    For j = 0 To NVI
        MatVarCovCoef(i, j) = s2 * A(i, j)
    Next j
Next i
'
' Nombre d'observations
'
nobs = NO + 1
'
' Degrs de libert
'
DoFree = nobs - NVI - 1
'
' Calcul de l'cart-type des coefficients et du t-student
'
For i = 0 To NVI
    '
    ' Coefficients
    '
    Coef(i, 0) = A(i, NC)
    '
    ' MexVal
    '
    Coef(i, 1) = 100 * (Sqr(1 + (A(i, NC) ^ 2 / (A(i, i) * A(NC, NC)))) - 1)
    Mexval(i) = 100 * (Sqr(1 + (A(i, NC) ^ 2 / (A(i, i) * A(NC, NC)))) - 1)
    '
    ' Elasticit par rapport  la moyenne
    '
    Coef(i, 2) = A(i, NC) * (Temp(i, 0) / Temp(NC, 0))
    Elasticity(i) = A(i, NC) * (Temp(i, 0) / Temp(NC, 0))
    '
    ' NorRes
    '
    Coef(i, 3) = TempS(i) / TempS(NVI)
    NorRes(i) = TempS(i) / TempS(NVI)
    '
    ' Ecart-Type du coefficient
    '
    Coef(i, 4) = Sqr(MatVarCovCoef(i, i))
    SigmaCoef(i) = Sqr(MatVarCovCoef(i, i))
    '
    ' Moyenne de la variable
    '
    If i = 0 Then
        Mean(i) = 1#
    Else
        Mean(i) = Temp(i, 0)
    End If
    Coef(i, 5) = Mean(i)
    '
    ' Beta
    '
    Coef(i, 6) = A(i, NC) * (Sqr(Temp(i, i)) / Sqr(Temp(NC, NC)))
    Beta(i) = A(i, NC) * (Sqr(Temp(i, i)) / Sqr(Temp(NC, NC)))
    '
    ' T de Student du coefficient
    '
    Coef(i, 7) = A(i, NC) / SigmaCoef(i)
    tstudent(i) = A(i, NC) / SigmaCoef(i)
    '
    ' Fisher
    '
    Coef(i, 8) = (R2 / (1 - R2)) * ((NOR - NVIR) / (NVIR - 1))
    FTest = (R2 / (1 - R2)) * ((NOR - NVIR) / (NVIR - 1))
Next i
'
' YHAT, valeur calcule de Y
'
For j = 0 To NVI
    For i = 0 To NO
        YHAT(i) = YHAT(i) + (A(j, NC) * X(i, j))
    Next i
    '
    ' Calcul des rsidus
    '
    For i = 0 To NO
        Resid(i) = X(i, NC) - YHAT(i)
    Next i
Next j
'
' MAPE
'
For i = 0 To NO
    MAPETEMP = MAPETEMP + Abs(Resid(i) / X(i, NC))
Next i
MAPE = 100 * (MAPETEMP / NOR)
'
' RHO et DW
'
For i = 0 To NO - 1
    DW = DW + (Resid(i) - Resid(i + 1)) ^ 2
Next i
DW = DW / SR2
RHO = (2 - DW) / 2
'
' SEE+1
'
lresid = 0#
For i = 0 To NO
    errp1 = Resid(i) - RHO * lresid
    somme1 = somme1 + errp1 * errp1
    lresid = Resid(i)
Next i
SEEP1 = Sqr(somme1 / (NO + 1))
'
' Impression "scalaire"
'
StatReg(0) = SEE
StatReg(1) = R2
StatReg(2) = RHO
StatReg(3) = nobs
StatReg(4) = 1
StatReg(5) = SEEP1
StatReg(6) = RBSQR
StatReg(7) = DW
StatReg(8) = DoFree
StatReg(9) = NO + 1
StatReg(10) = MAPE
'
' 0   : les numros des observations
' 1   : Les observations de Y
' 2   : Les observations de Yhat
' 3   : Les observations des rsidus
'
For i = 0 To UBound(X, 1)
    RegData(i, 0) = i + 1
    RegData(i, 1) = X(i, NC)
    RegData(i, 2) = YHAT(i)
    RegData(i, 3) = Resid(i)
    'Debug.Print RegData(i, 0) & " " & RegData(i, 1) & " " & RegData(i, 2) & " " & RegData(i, 3)
Next i

'Debug.Print "MeanY = " & Temp(NC, 0)
'
'Debug.Print "Sigma  = " & s2
'
'Debug.Print "SigmaY = " & Sqr(Temp(NC, NC) / NOR)
'
'----------------------------------------------------------------------------------------
' Fin du sous-programme de rgression.
'----------------------------------------------------------------------------------------
'
End Sub



Public Function Variance(X() As Double, Mean() As Double) As Double()
'
'
'
'
Dim SommeCarre() As Double
Dim Var() As Double
Dim i As Long
Dim j As Long
'Dim Mean() As Double
'
'
'
ReDim Moyenne(0 To UBound(X, 2))
ReDim SommeCarre(0 To UBound(X, 2))
ReDim Variance(0 To UBound(X, 2))
ReDim Var(0 To UBound(X, 2))
'ReDim Mean(0 To UBound(X, 2))
'
'
'
'Mean = Moyenne(X())
For i = 0 To UBound(X, 2)
'    Debug.Print "Mean = " & Mean(i)
Next i

For i = 0 To UBound(X, 2)
    For j = 0 To UBound(X, 1)
        SommeCarre(i) = SommeCarre(i) + (X(j, i) - Mean(i)) ^ 2
        'Debug.Print "SommeCarre(" & i & ") = " & SommeCarre(i)
    Next j
    Var(i) = SommeCarre(i) / (UBound(X, 1) + 1)
'    Debug.Print "Variance(" & i & ")= " & Var(i)
'    Debug.Print "SD(" & i & ")= " & Sqr(Var(i))
Next i
'
'
'
Variance = Var
'
'
'
End Function

Public Function Moyenne(X() As Double) As Double()
'
'
'
Dim Somme() As Double
Dim Mean()  As Double
Dim i As Long
Dim j As Long
'
'
'
ReDim Mean(0 To UBound(X, 1))
ReDim Moyenne(0 To UBound(X, 1))
ReDim Somme(0 To UBound(X, 1))
'
'
'
For j = 0 To UBound(X, 2)
    Somme(0) = 0#
    For i = 0 To UBound(X, 1)
        Somme(j) = Somme(j) + X(i, j)
        'Debug.Print "i = " & i & " j = " & j & " X = " & X(i, j)
    Next i
    Mean(j) = Somme(j) / (UBound(X, 1) + 1)
'    Debug.Print "Moyenne(" & j & ")= " & Mean(j)
Next j
'
'
'
Moyenne = Mean
'
'
'
End Function

Public Function EcartType(Variance() As Double) As Double()
'
'
'
Dim i As Long
Dim ET() As Double
'
'
'
ReDim EcartType(0 To UBound(Variance))
ReDim ET(0 To UBound(Variance))
'
'
'
For i = 0 To UBound(Variance)
    ET(i) = Sqr(Variance(i))
'    Debug.Print "Ecart-Type(" & i & ") = " & ET(i)
Next i
'
EcartType = ET
'
'
'
End Function

Public Function CoVariance(X() As Double, Mean() As Double) As Double()
'
'
'
'
Dim CovVar() As Double
Dim Var() As Double
Dim i As Long
Dim j As Long
Dim nobs As Long
Dim NC As Long '
Dim NO As Long
'
'
'
NC = UBound(X, 2)
NO = UBound(X, 1)
'
'
'
ReDim Moyenne(0 To NC)
ReDim CoVariance(0 To NC, 0 To NC)
ReDim CovVar(0 To NC, 0 To NC)
'
'
'
For i = 0 To NC
    For j = 0 To NC
        CovVar(i, j) = 0#
        For nobs = 0 To NO
            CovVar(i, j) = CovVar(i, j) + (X(nobs, i) - Mean(i)) * (X(nobs, j) - Mean(j))
        Next nobs
        CovVar(i, j) = CovVar(i, j) / nobs
    Next j
Next i
'
'
'
CoVariance = CovVar
'
'
'
End Function

